{- 2010-09-17 haskell syntax tools -}

module HsSyntax where

import Language.Haskell.Exts.Syntax

noLoc :: SrcLoc
noLoc = SrcLoc "" 0 0

qualIO :: String -> QName
qualIO s = Qual (ModuleName "IO") (Ident s)

mkModule :: [Decl] -> Module
mkModule hs = Module noLoc main_mod pragmas warning exports imports decls where
  pragmas = [ LanguagePragma noLoc
    [ Ident "NoImplicitPrelude"
    ]]
  warning = Nothing
  exports = Nothing
  imports = 
    [ ImportDecl 
      { importLoc       = noLoc 
      , importModule    = ModuleName "System.IO" 
      , importQualified = True 
      , importSrc       = False 
      , importPkg       = Nothing
      , importAs        = Just $ ModuleName "IO"
      , importSpecs     = Nothing
      } 
    ]
  decls   = hs ++ 
    [ TypeSig noLoc [ main_name ] io
    , FunBind [ Match noLoc main_name [] Nothing (UnGuardedRhs rhs) (BDecls []) ]
    ] where rhs  = Var (qualIO "putStrLn") `App` Lit (String "Hello, world!")
            io   = TyCon (qualIO "IO") `TyApp` unit_tycon
         
noContext = []
noDeriving = []
noTyVarBind = []

mkDataDecl :: Name -> [TyVarBind] -> Kind -> [GadtDecl] -> Decl
mkDataDecl n tel k cs = GDataDecl noLoc DataType noContext n tel (Just k) cs noDeriving

mkConDecl :: Name -> Type -> GadtDecl
mkConDecl n t = GadtDecl noLoc n t

mkTyFun :: Type -> Type -> Type
mkTyFun a b = mkTyParen a `TyFun` b

mkForall :: Name -> Kind -> Type -> Type
mkForall x k t = TyForall (Just $ [KindedVar x k]) noContext t

mkTyParen :: Type -> Type
mkTyParen a@(TyVar{}) = a
mkTyParen a@(TyCon{}) = a
mkTyParen a = TyParen a

mkTyApp :: Type -> Type -> Type
mkTyApp f a = TyApp f a

noBinds = BDecls []

mkTypeSig :: Name -> Type -> Decl
mkTypeSig x t = TypeSig noLoc [x] t

-- create a simple function clause x = t
mkLet :: Name -> Exp -> Decl
mkLet x e = FunBind [Match noLoc x [] Nothing (UnGuardedRhs e) noBinds]

mkVar :: Name -> Exp
mkVar = Var . UnQual

mkLam :: Name -> Exp -> Exp
mkLam x (Lambda _ ps e) = Lambda noLoc (PVar x : ps) e
mkLam x  e              = Lambda noLoc [PVar x] e

mkParen :: Exp -> Exp
mkParen e@(Var{}) = e
mkParen e@(Con{}) = e
mkParen e = Paren e

mkApp :: Exp -> Exp -> Exp
mkApp f e = App f (mkParen e) 

{- this is already predefined as unit_con
hsDummyExp :: HsExp
hsDummyExp = HsCon $ Special $ HsUnitCon  -- Haskell's '()'
-}