{ 
module Parser where

import qualified Lexer as T
import qualified Concrete as C

import Abstract (Decoration(..),Dec,defaultDec,Override(..))
import Polarity (Pos(..))
import qualified Abstract as A
import qualified Polarity as A
import Abstract (Name)
}

%name parse
%tokentype { T.Token }
%error { parseError }

%token

id      { T.Id $$ _ }
number  { T.Number $$ _ }
data    { T.Data _ }
codata  { T.CoData _ }
sized   { T.Sized _ }
fields  { T.Fields _ }
mutual  { T.Mutual _ }
fun     { T.Fun _ }
cofun   { T.CoFun _ } 
case    { T.Case _ }
let     { T.Let _ }
in      { T.In _ }
eval    { T.Eval _ }
fail    { T.Fail _ }
check   { T.Check _ }
trustme { T.TrustMe _ }
impredicative { T.Impredicative _ }
type    { T.Type _ }
set     { T.Set _ }
coset   { T.CoSet _ }
size    { T.Size _ }
infty   { T.Infty _ }
succ    { T.Succ _ }
max     { T.Max _ }
'<'     { T.AngleOpen _ }
'>'     { T.AngleClose _ }
'{'     { T.BrOpen _ }
'}'     { T.BrClose _ }
'['     { T.BracketOpen _ }
']'     { T.BracketClose _ }
'('     { T.PrOpen _ }
')'     { T.PrClose _ }
'|'     { T.Bar _ }
','     { T.Comma _ }
';'     { T.Sem _ }
':'     { T.Col _ }
'.'     { T.Dot _ }
'->'    { T.Arrow _ }
'='     { T.Eq _ }
'++'    { T.PlusPlus _ }
'+'     { T.Plus _ }
'-'     { T.Minus _ }
'/'     { T.Slash _ }
'*'     { T.Times _ } -- UNUSED
'\\'    { T.Lam _ }
'_'     { T.Underscore _ }

%%

TopLevel :: { [C.Declaration] }
TopLevel : Declarations { reverse $1}


Declarations :: { [C.Declaration] }
Declarations : {- empty -} { [] }
             | Declarations Declaration { $2 : $1 }

Declaration :: { C.Declaration }
Declaration : Data                      { $1 }
           | CoData                     { $1 }
           | SizedData                  { $1 }
           | SizedCoData                { $1 }
           | Fun                        { $1 }
           | CoFun                      { $1 }
           | Mutual                     { $1 }
           | Let                        { $1 }
           | impredicative Declaration          { C.OverrideDecl Impredicative [$2] }
           | impredicative '{' Declarations '}' { C.OverrideDecl Impredicative $3 }
           | fail Declaration             { C.OverrideDecl Fail [$2] }
           | fail '{' Declarations '}'    { C.OverrideDecl Fail $3 }
           | check Declaration            { C.OverrideDecl Check [$2] }
           | check '{' Declarations '}'   { C.OverrideDecl Check $3 }
           | trustme Declaration          { C.OverrideDecl TrustMe [$2] }
           | trustme '{' Declarations '}' { C.OverrideDecl TrustMe $3 }

Data :: { C.Declaration }
Data : data Id DataTelescope ':' Expr '{' Constructors '}' OptFields
   { C.DataDecl $2 A.NotSized A.Ind $3 $5 (reverse $7) $9 }

SizedData :: { C.Declaration }
SizedData : sized data Id DataTelescope ':' Expr '{' Constructors '}' OptFields
   { C.DataDecl $3 A.Sized A.Ind $4 $6 (reverse $8) $10 }

CoData :: { C.Declaration }
CoData : codata Id DataTelescope ':' Expr '{' Constructors '}' OptFields 
       { C.DataDecl $2 A.NotSized A.CoInd $3 $5 (reverse $7) $9 }

SizedCoData :: { C.Declaration }
SizedCoData : sized codata Id DataTelescope ':' Expr '{' Constructors '}' OptFields
       { C.DataDecl $3 A.Sized A.CoInd $4 $6 (reverse $8) $10 }

Fun :: { C.Declaration }
Fun : fun TypeSig '{' Clauses '}' { C.FunDecl A.Ind $2 $4 }

CoFun :: { C.Declaration }
CoFun : cofun TypeSig '{' Clauses '}' { C.FunDecl A.CoInd $2 $4  }



Mutual :: { C.Declaration }
Mutual : mutual '{' Declarations '}' { C.MutualDecl (reverse $3) }
     

 
Let :: { C.Declaration }
Let : let TypeSig '=' Expr { C.LetDecl False $2 $4 } 
      | eval let TypeSig '=' Expr { C.LetDecl True $3 $5 } 

OptFields :: { [Name] }
OptFields : {- empty -}  { [] }
          | fields Ids   { $2 }
-----

Id :: { Name }
Id : id { $1 }
-- no longer  number { $1 }

SpcIds :: { [Name] } -- non-empty list
SpcIds : Id     { [$1] }
       | Id SpcIds { $1 : $2 }

Ids :: { [Name] } -- non-empty list
Ids : Id              { [$1] }
    | Id ',' Ids { $1 : $3 }

Pos :: { Pos }
Pos : '++'         { SPos  }
    | '+'          { Pos   }
    | '-'          { Neg   }
    | '/'          { NOcc  }
--    | {- empty -}  { Mixed }

Measure :: { A.Measure C.Expr }
Measure : '|' Meas { A.Measure $2 }

Meas :: { [C.Expr] }
Meas : Expr '|'      { [$1] }
     | Expr ',' Meas { $1 : $3 }

Bound :: { A.Bound C.Expr }
Bound : Measure '<' Measure { A.Bound $1 $3 }

TBind :: { C.TBind }
TBind :  '(' Ids ':' Expr ')' { C.TBind (Dec False Mixed) $2 $4 } -- ordinary binding
      |  '[' Ids ':' Expr ']' { C.TBind (Dec True Mixed) $2 $4 }  -- erased binding
      |  Pos '(' Ids ':' Expr ')' { C.TBind (Dec False $1) $3 $5 } -- ordinary binding
      |  Pos '[' Ids ':' Expr ']' { C.TBind (Dec True $1) $3 $5 }  -- erased binding

-- let binding
LBind :: { C.TBind }
LBind :  Id ':' Expr         { C.TBind (Dec False Mixed) [$1] $3 } -- ordinary binding
      |  '(' Id ':' Expr ')' { C.TBind (Dec False Mixed) [$2] $4 } -- ordinary binding
      |  '[' Id ':' Expr ']' { C.TBind (Dec True Mixed) [$2] $4 }  -- erased binding
      |  Pos '(' Id ':' Expr ')' { C.TBind (Dec False $1) [$3] $5 } -- ordinary binding
      |  Pos '[' Id ':' Expr ']' { C.TBind (Dec True $1) [$3] $5 }  -- erased binding


Domain :: { C.TBind }
Domain : Expr1             { C.TBind (Dec False Mixed) [] $1 }
       | '[' Expr ']'      { C.TBind (Dec True  Mixed) [] $2 }
       | Pos Expr1         { C.TBind (Dec False $1) [] $2 }
       | Pos '[' Expr ']'  { C.TBind (Dec True  $1) [] $3 }
       | TBind             { $1 }
       | Measure           { C.TMeasure $1 }
       | Bound             { C.TBound $1 }

-- general form of expression
Expr :: { C.Expr }
Expr : Domain '->' Expr                   { C.Pi $1 $3 } 
       | '\\' SpcIds '->' Expr               { foldr C.Lam $4 $2 }
       | let LBind '=' Expr in Expr       { C.LLet $2 $4 $6 }
       | case Expr '{' Cases '}'          { C.Case $2 $4 }  
       | Expr1                            { $1 }

-- perform applications
Expr1 :: { C.Expr }
Expr1 : Expr2 { let list = reverse $1 in
               if length list == 1 then head list else C.App (head list) (tail list) 
	      }
       | coset Expr3                      { C.CoSet $2 }
       | set                              { C.Set C.Zero }
       | set Expr3                        { C.Set $2 }

-- gather applications
Expr2 :: { [C.Expr] }
Expr2 : Expr3 { [$1] }
       | Expr2 Expr3 { $2 : $1 }
       | Expr2 set   { C.Set C.Zero : $1 }
--       | succ SE { [C.Succ $2] }

-- atoms
Expr3 :: { C.Expr }
Expr3 : size                   { C.Size }
      | max                    { C.Max }
      | infty                  { C.Infty }
      | Id                     { C.Ident $1}
      | '<' Expr ':' Expr '>'  { C.Sing $2 $4 }
      | '(' Expr ')'           { $2 }
      | '_'                    { C.Unknown }
      | succ Expr3             { C.Succ $2 }  -- succ is a prefix op
      | number                 { iterate C.Succ C.Zero !! (read $1) }

{- RETIRED
SE :: { C.Expr}
SE : succ SE { C.Succ $2}
SE : Expr3 { $1}
  -}

TypeSig :: { C.TypeSig }
TypeSig : Id ':' Expr { C.TypeSig $1 $3 }

Constructor :: { C.Constructor }
Constructor : TypeSig { $1 }

Constructors :: { [C.Constructor ] }
Constructors :
      Constructors ';' Constructor { $3 : $1 }
    | Constructors ';' { $1 }
    | Constructor { [$1] }
    | {- empty -} { [] }

Cases :: { [C.Clause] }
Cases : Pattern '->' Expr ';' Cases  { (C.Clause Nothing [$1] (Just $3)) : $5 }
      | Pattern '->' Expr            { (C.Clause Nothing [$1] (Just $3)) : [] }
      | Pattern ';' Cases            { (C.Clause Nothing [$1] Nothing) : $3 }
      | Pattern                      { (C.Clause Nothing [$1] Nothing) : [] }
      | {- empty -}                  { [] }
      
Clause :: { C.Clause }
Clause : Id LHS '=' Expr { C.Clause (Just $1) $2 (Just $4) }
       | Id LHS          { C.Clause (Just $1) $2 Nothing }

LHS :: { [C.Pattern] }
LHS : Patterns { reverse $1 }

Patterns :: { [C.Pattern] }
Patterns : {- empty -} { [] }
    | Patterns Pattern { $2 : $1 }

Pattern :: { C.Pattern }
Pattern : ConP               { $1            }
        | Id                 { C.IdentP $1   }
        | '.' set            { C.DotP (C.Set C.Zero) }
        | '.' Expr3          { C.DotP $2     }
        | '(' Id '>' Id ')'  { C.SizeP $2 $4 } 
--        | '*'       { C.AbsurdP }

ConP :: { C.Pattern }
ConP : '(' Id Patterns ')' { C.ConP $2 (reverse $3) }
     | '(' succ SP ')'     { C.SuccP $3             }
     | '(' ')'             { C.AbsurdP              }

SP :: { C.Pattern}
SP : succ SP {C.SuccP $2}
   | Pattern { $1 }

Clauses :: { [C.Clause] }
Clauses : RClauses { reverse $1 }

RClauses :: { [C.Clause ] }
RClauses :
   RClauses ';' Clause { $3 : $1 }
 | RClauses ';' { $1 }
 | Clause { [$1] }
 | {- empty -} { [] }

-- for backwards compatibility
TBindSP :: { C.TBind }
TBindSP : '(' Ids ':' Expr ')' { C.TBind (Dec False Mixed) $2 $4 } -- ordinary binding
        | '[' Ids ':' Expr ']' { C.TBind (Dec True Mixed) $2 $4 }  -- erased binding
        | Pos '(' Ids ':' Expr ')' { C.TBind (Dec False $1) $3 $5 } 
        | Pos '[' Ids ':' Expr ']' { C.TBind (Dec True $1) $3 $5 }  
        |  '(' '+' Ids ':' Expr ')' { C.TBind (Dec False SPos) $3 $5 }
        |  '[' '+' Ids ':' Expr ']' { C.TBind (Dec True SPos) $3 $5 }

DataTelescope :: { C.Telescope }
DataTelescope :  {- empty -} { [] }
--                | TBind   DataTelescope { $1 : $2 } 
                | TBindSP DataTelescope { $1 : $2 } 

{

parseError :: [T.Token] -> a
parseError [] = error "Parse error at EOF"
parseError (x : xs) = error ("Parse error at token " ++ T.prettyTok x) 

}

 



